//  Copyright 2009 David Ekholm, Jalbum AB
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.package net.jalbum.editor;

package net.jalbum.diskusage;


import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.text.DecimalFormat;
import java.util.Date;
import java.util.StringTokenizer;

/**
 *
 * @author david
 */
public class DiskUsage {

    public static final String CACHE_FILE_NAME = ".du";
    private static final DecimalFormat twoDec = new DecimalFormat("0.##");

    private boolean verbose = false;
    private int nWrites = 0;
    private int nDirs = 0;
    private boolean forceCacheRefresh = false;

    public boolean isForceCacheRefresh() {
        return forceCacheRefresh;
    }

    public void setForceCacheRefresh(boolean forceCacheRefresh) {
        this.forceCacheRefresh = forceCacheRefresh;
    }

    public int getNDirs() {
        return nDirs;
    }

    public void setNDirs(int nDirs) {
        this.nDirs = nDirs;
    }

    public int getNWrites() {
        return nWrites;
    }

    public void setNWrites(int nWrites) {
        this.nWrites = nWrites;
    }

    public boolean isVerbose() {
        return verbose;
    }

    public void setVerbose(boolean verbose) {
        this.verbose = verbose;
    }

    public DiskUsage() {
    }

    public static long getDiskUsage(File path) throws IOException {
        return getDiskUsage(path, false);
    }

    public static long getDiskUsage(File path, boolean forceCacheRefresh) throws IOException {
        DiskUsage du = new DiskUsage();
        du.setForceCacheRefresh(forceCacheRefresh);
        return du.getUsage(path);
    }

    public long getUsage(File path) throws IOException {
        nDirs++;
        File cacheFile = new File(path, CACHE_FILE_NAME);
        if (forceCacheRefresh  || !cacheFile.exists() || path.lastModified() > cacheFile.lastModified()) {
            if (verbose) {
                if (!cacheFile.exists()) {
                    System.out.println("Missing cache file: " + cacheFile);
                }
                else {
                    System.out.println("Path " + path + " last modified: "
                            + new Date(path.lastModified())
                            + ". Cache file last modified: " + new Date(cacheFile.lastModified()));
                }
            }
            writeCacheFile(path);
        }
        String content = readTextFile(cacheFile);
        StringTokenizer entries = new StringTokenizer(content, "\n");
        long usage = Long.parseLong(entries.nextToken());
        while (entries.hasMoreTokens()) {
            File sub = new File(path, entries.nextToken());
            if (!sub.exists()) {
                System.err.println("DiskUsage: Can't find referenced path: " + sub + ". Refreshing cache file");
                cacheFile.delete();
                return getDiskUsage(path); // Re-run this directory
            }
            usage += getUsage(sub);
        }
        return usage;
    }

    private static String formatAsMB(long usage) {
        return twoDec.format(((double)usage)/1024/1024);
    }

    private void writeCacheFile(File dir) throws IOException {
        StringBuilder buf = new StringBuilder();
        long usage = 0;
        File[] files = dir.listFiles();
        if (files == null) {
            throw new RuntimeException("Invalid directory: " + dir.getAbsolutePath());
        }
        for (File f : files) {
            if (f.isDirectory()) {
                buf.append(f.getName());
                buf.append('\n');
            } else {
                if (!f.getName().equals(CACHE_FILE_NAME)) {
                    usage += f.length();
                }
            }
        }
        String content = "" + usage + '\n' + buf.toString();
        File manifest = new File(dir, CACHE_FILE_NAME);
        writeTextFile(content, manifest);
    }

    public static void deleteCacheFiles(File path) {
        new File(path, CACHE_FILE_NAME).delete();
        File[] files = path.listFiles(new FileFilter() {
            public boolean accept(File pathname) {
                return pathname.isDirectory();
            }
        });
        for (File f : files) {
            deleteCacheFiles(f);
        }
    }

    private static String readTextFile(File f) throws IOException {
        byte[] bytes = readBytes(f);
        return new String(bytes, "UTF-8");
    }

    private static byte[] readBytes(File f) throws IOException {
        byte[] buf = new byte[(int)f.length()];
        BufferedInputStream in = new BufferedInputStream(new FileInputStream(f));
        in.read(buf);
        in.close();

        return buf;
    }

    private void writeTextFile(String content, File file)
        throws IOException {
        BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), "UTF-8"));
        writer.write(content);
        writer.close();
        nWrites++;
    }

    public static void main(String[] args) throws IOException {
        if (args.length == 0) {
            System.out.println("Usage: DiskUsage [-f | -d | -v] [<path>]");
            return;
        }
        boolean forceRefresh = args[0].equals("-f");
        boolean deleteCache = args[0].equals("-d");
        boolean verbose = args[0].equals("-v");

        File path = new File("").getAbsoluteFile();
        if (!args[args.length-1].startsWith("-")) {
            path = new File(args[args.length-1]).getAbsoluteFile();
        }
        if (deleteCache) {
            deleteCacheFiles(path);
            System.out.println("Disk usage cache files under " + path + " deleted");
        }
        else {
            DiskUsage du = new DiskUsage();
            du.setVerbose(verbose);
            du.setForceCacheRefresh(forceRefresh);
            long usage = du.getUsage(path);
            String stats = ". Dirs processed: " + du.getNDirs() + ". Cache writes: " + du.getNWrites();
            System.out.println("Disk usage for " + path + ": " + usage + " bytes (" + formatAsMB(usage) + " MB)" + stats);
        }

    }

}
